r"""
DIRAC (Domain Invariant Representation through Adversarial Calibration)
"""

from ._version import __version__
from .adj import *
from .dataprep import *
from .utils import *
from .hyper import *
from .main import *
from .model import *
from .trainer import *
